(***********************************************************************)
(*                                                                     *)
(*                          Sklml                                      *)
(*                                                                     *)
(*             Pierre Weis (Pierre.Weis@inria.fr)                      *)
(*             Francois Clement (Francois.Clement@inria.fr)            *)
(*             Quentin Carbonneaux (Quentin.Carbonneaux@inria.fr)      *)
(*                                                                     *)
(* This program is free software.                                      *)
(* (see the file LICENSE in the root directory of the distribution).   *)
(*                                                                     *)
(***********************************************************************)

(* $Id: skl.ml,v 1.3 2010-10-22 07:07:36 weis Exp $ *)

type ('a, 'b) skel = 'a Stream.t -> 'b Stream.t;;
type ('a, 'b) skel_init = unit -> ('a, 'b) skel;;

let message s = Printf.eprintf "%s\n" s;;

let skel_of_fun mess f =
  message (Printf.sprintf "%s: creating a skeleton from a function" mess);
  f
;;

let make fs = skel_of_fun "Skl.make" fs;;

let fun_of_skel mess skl =
  message (Printf.sprintf "%s: creating a function from a skeleton" mess);
  skl
;;

let from skl = fun_of_skel "Skl.from" skl;;

let initialize_skel mess f =
  message (Printf.sprintf "%s: initializing a skeleton" mess);
  f ()
;;

let parfun_mess mess skl =
  let f = fun_of_skel mess skl in
  fun s -> f s
;;

let parfun skl = parfun_mess "parfun" skl
;;

let lift f = skel_of_fun "lift" (Stream.map f)
;;

let unlift skl = Stream.unlift (from skl);;

let lift_init fi () =
  let f = initialize_skel "lift_init" fi in
  lift f
;;

let unlift_init skli =
  let skl = initialize_skel "unlift_init" skli in
  unlift skl
;;

(*let seq_apply of_stream skl to_stream x = of_stream (skl (to_stream x));;*)
(** Turns a skeleton from a function from lazy stream of A to lazy stream of B
   into a regular function from A to B, by creating a stream from the
   regular argument and extracting a regular value from the resulting
   stream. *)

(*let unseq_init skli =
  let skl = initialize_skel "unseq_init" skli in
  let f = parfun_mess "unseq" skl in
  fun x -> Stream.head (f (Stream.singleton x))
;;

let unseq skl = unseq_init (fun () -> skl);;
*)
let pipeline skl1 skl2 =
  let f1 = fun_of_skel "pipe: first skeleton" skl1 in
  let f2 = fun_of_skel "pipe: second skeleton" skl2 in
  fun s -> f2 (f1 s)
;;
(* Should write something like
   let c1, s1 = get_cell f1 s in
   let c2 = f2 (c1) in
   ...
*)

let ( ||| ) = pipeline;;

let par_apply skl1 skl2 =
  let f1 = fun_of_skel "par_apply: first skeleton" skl1 in
  let f2 = fun_of_skel "par_apply: second skeleton" skl2 in
  function (s1, s2) -> (f1 s1, f2 s2)
;;
(** Returns a parallel circuit that applies 2 skeletons in parallel
  onto 2 streams. *)

let in_farm collect (skl, n) =
  if n <= 0 then failwith "farm: non positive number of workers." else
  let rec in_farm n =
    if n = 1 then initialize_skel "in_farm" skl else
    let r = n mod 2
    and q = n / 2 in
    collect
      (par_apply
        (in_farm (q + r))
        (in_farm q)) in
  in_farm n
;;

let do_plug f c s = f (c (Stream.fork s));;
(** [do_plug f c s] plugs a parallel circuit [c] with 2 inputs and 2 outputs
   to 2 wires: one wire for input with a [fork] emitter to the two input
   wires of [c], and one wire for output with a [f] collector to the two
   output wires of [c]). *)

let plug c = do_plug Stream.join c;;
(** [plug c] returns a circuit that fork a stream [s] and applies the parallel
   circuit [c] to those 2 streams, then joins the results into a resulting
   stream.
   In short, [plug] plugs a parallel circuit with 2 inputs and 2 outputs to
   2 wires (with an emitter and a collector at both ends).
   [plug c s] = [join (c (fork s))]. *)

let farm_init (skli, n) = in_farm plug (skli, n);;

let farm (skl, n) = farm_init ((fun () -> skl), n);;

let farm_vector_init (skli, n) =
  let fskl = farm_init (skli, n) in
  Stream.serialize_stream_vector
  ||| fskl |||
  Stream.unserialize_stream_vector
;;

let farm_vector (skl, n) = farm_vector_init ((fun () -> skl), n);;

let seq_reduce (f : 'a * 'b -> 'a) (z : 'a) (s : 'b Stream.stream) =
  Stream.fold_left f z s;;

let reduce_init ((skli : unit -> ('a, 'b) skel), n) z =
  if n <= 0 then
    failwith "reducevector: non positive number of workers." else
  let rskli () =
    let f = unlift_init skli in
    seq_reduce f z in
  let rfarm = farm_init (rskli, n) in
  let f = unseq_init skli in
  rfarm
;;

let reduce (skl, n) = reduce_init ((fun () -> skl), n);;

let reduce_vector_init (skli, n) z =
  let rskl = reduce_init (skli, n) z in
  Stream.serialize_stream_vector ||| rskl
;;

let reduce_vector (skl, n) = reduce_vector_init ((fun () -> skl), n);;

let loop (p, skl) s =
  let f = unlift skl in
  let rec fix_f x =
    let y = f x in
    if p y then fix_f y else y in
  Stream.map fix_f s
;;

let pardo f =
  message "Launching the computational network";
  f ()
;;

let seq = lift;;
let mapvector = farm_vector;;
let reducevector = reduce_vector;;

(*
val skel_of_fun : ('a Streams_wire.t -> 'b Streams_wire.t) -> ('a, 'b) skel;;
(** The private skeleton maker. *)

val fun_of_skel : ('a, 'b) skel -> ('a Stream.t -> 'b Stream.t)
(** The private function from skeleton maker. *)
;;
*)
